package DFA::Cellular;
use Curses;

sub new {
	my ($class, $width, $height, $init) = @_;
	$init = [ map {[split '', substr($_,0,$width)]} split "\n", $init, $height ];
	return bless {	cstate => $init,
			width  => $width,
			height => $height,
			rules  => [],
		      },
		      $class;
}

sub addrule {
	my ($self, @rules) = @_;
	push @{$self->{rules}}, @rules;
}

sub step {
	my ($self) = @_;
	$self->{nstate} = [map {[@$_]} @{$self->{cstate}}];
	my ($x, $y, $rule);
	for $x (1..$self->{width}-2) {
	    for $y (1..$self->{height}-2) {
		my $neigh = $self->cneigh($x,$y);
		for $rule ( @{$self->{rules}} ) {
		    my ($nstate) = $rule->transforms($neigh);
		    next unless defined $nstate;
		    $self->{nstate}[$y][$x]=$nstate;
		    last;
		}
	    }
	}
	$self->{cstate} = $self->{nstate};
}

sub show {
	my ($self) = @_;
	$self->{show} ||= Curses->new();
	$self->{show}->addstr(0,0, $self);
	$self->{show}->refresh;
}

sub cneigh {
	my ($self, $x, $y) = @_;
	my @x = ($x-1..$x+1);
	return [ [@{$self->{cstate}[$y-1]}[@x]],
	         [@{$self->{cstate}[ $y ]}[@x]],
	         [@{$self->{cstate}[$y+1]}[@x]],
	       ];
}

sub mapout {
	my ($self, $mapper) = @_;
	$self->{mapper} = $mapper;
}

use overload
  q{""} => sub {
	my $mapper = $_[0]->{mapper};
	if ($mapper) {
		join "\n", map {join "", map $mapper->($_), @$_}
				@{$_[0]->{cstate}},""
	}
	else {
		join "\n", map {join "", @$_} @{$_[0]->{cstate}},""
	}
  };


package DFA::Cellular::Rule;
use Carp;

sub new {
	my ($class, $defn) = @_;
	croak "Expected rule specification via array reference"
		unless ref($defn) eq 'ARRAY';
	$from = [map { [split "", substr($_,0,3)] } @{$defn}[0,1,3]];
	$to   = substr($defn->[2],0,1);

	bless { from => $from, to => $to }, $class;
}

sub debug {
	my ($what, $self, $neigh) = @_;
	print "$what\n\t",
		join("/", map {join "", @$_} @{$self->{from}}),
		"\n\t",
		join("/", map {join "", @$_} @{$neigh}),
		"\n"
}


sub transforms {
	my ($self, $neigh) = @_;
	# debug "Trying", $self, $neigh;
	for $x (0..2) {
	    for $y (0..2) {
	        my $cell = $self->{from}[$y][$x];
	        return unless $neigh->[$y][$x] eq $cell || $cell eq '.'
	    }
	}
	# debug ">>>>>>", $self, $neigh;
	return $self->{to};
}


sub mirror {
	my ($self) = @_;
	my $type = ref $self;
	my $hor = $type->new(["   ","   ",$self->{to},"   "]);
	my $ver = $type->new(["   ","   ",$self->{to},"   "]);
	my $hav = $type->new(["   ","   ",$self->{to},"   "]);
	my ($x, $y);
	for $x (0..2) {
	    for $y (0..2) {
		$hor->{from}[$y][$x] = $self->{from}[ $y ][2-$x];
		$ver->{from}[$y][$x] = $self->{from}[2-$y][ $x ];
		$hav->{from}[$y][$x] = $self->{from}[2-$y][2-$x];
	    }
	}
	return ($self, $hor, $ver, $hav);
}

sub rotate {
	my ($self) = @_;
	my $type = ref $self;
	my $r090 = $type->new(["   ","   ",$self->{to},"   "]);
	my $r180 = $type->new(["   ","   ",$self->{to},"   "]);
	my $r270 = $type->new(["   ","   ",$self->{to},"   "]);
	my ($x, $y);
	for $x (0..2) {
	    for $y (0..2) {
		$r090->{from}[$y][$x] = $self->{from}[ $x ][2-$y];
		$r180->{from}[$y][$x] = $self->{from}[2-$y][2-$x];
		$r270->{from}[$y][$x] = $self->{from}[2-$x][ $y ];
	    }
	}
	return ($self, $r090, $r180, $r270);
}


package DFA::Cellular::ProcRule;

sub new {
	my ($class, $proc) = @_;

	bless { proc => $proc }, $class;
}

sub transforms {
	my ($self, $neigh) = @_;
	return $self->{proc}->($neigh);
}

1;
__END__

